/*
 * inertia.hh
 *
 *  Created on: May 11, 2010
 *      Author: mauro
 */

#ifndef MLN_ACCU_SHAPE_INERTIA2D_HH
# define MLN_ACCU_SHAPE_INERTIA2D_HH

/// \file
/// Define an accumulator that computes the volume of a
/// component through one of its pixels.

/*!
This accumulator uses an mln::util::pix (pixel) to update the
reference level, area and volume information of the component.

The class mln/accu/volume is not a general-purpose accumulator;
it is used to implement volume-based connected filters.
\see mln::morpho::closing::volume
\see mln::morpho::opening::volume
 */

# include <mln/accu/internal/base.hh>
# include <mln/core/concept/meta_accumulator.hh>
# include <mln/math/diff_abs.hh>

# include <mln/util/pix.hh>
# include <mln/literal/zero.hh>

namespace mln
{

namespace accu
{

namespace shape
{

/// \brief Moment of Inertia accumulator class.
///
/// First moment invariant of Hu.
/// The parameter \p I is the image type on which the accumulator
/// of pixels is built.
///
/// \ingroup modaccuimages

/// \brief Generic bounding box accumulator class.
 ///
 /// The parameter \c P is the type of points.
 ///
 /// \ingroup modaccusiteset
 //

// TODO: Add the template of the image function.
template <typename P>
struct inertia2d : public mln::accu::internal::base< float , inertia2d<P> >
{
//	typedef mln_coord(P) argument;
	typedef P argument;
//	typedef p_array<mln_sum(mln_coord(P))> Ppow;
	typedef algebra::vec<P::dim, mln_sum(mln_coord(P))> Ppow;

//	typedef typename P::coord argument;
//	typedef P argument;

//	enum{dim=2};
//

//template <typename T,
//typename S = mln_sum(mln_psite(T)),
//typename R = S>
//struct inertia2d : public mln::accu::internal::base< R , inertia2d<T, S, R> >{
//
//	typedef mln_site(T) argument;
//	typedef R result;
////	typedef util::pix<I> argument;
//	typedef mln_site(I) argument;
//	typedef mln_sum(argument) 			p_array_pow1;
//	typedef mln_sum(mln_sum(argument)) 	p_array_pow2;
//	typedef util::pix<I> argument;
//	typedef typename argument::value value;

	inertia2d();

	/// Manipulators.
	/// \{
	void init();
	void take(const argument& pixel);
	void take(const inertia2d<P>& other);

	/// Force the value of the counter to \a v.
	void set_value(unsigned v);
	/// \}

	/// Get the value of the accumulator.
	float to_result() const;

	/// Check whether this accu is able to return a result.
	/// Always true here.
	bool is_valid() const;

protected:
//	/// The reference level (the level of the component's root).
//	value ref_level__;
//	/// The area of the component.
//	unsigned area__;
//	/// The inertia of the component.
//	unsigned inertia_;
//
//	p_array_pow1 sum_;
//	p_array_pow2 sum2_;

	unsigned area_;
	Ppow sum_;
//	P sum_;
	Ppow sum2_;
//	mln_vec(P) sum_;
//	mln_vec(P) sum2_;

};


} // end of mln::accu::shape


namespace meta
{

namespace shape
{

/// Meta accumulator for inertia.

struct inertia2d : public Meta_Accumulator< inertia2d >
{
	template <typename P>
	struct with
	{
		typedef accu::shape::inertia2d<P> ret;
	};
};

} // end of namespace mln::accu::meta::shape

} // end of namespace mln::accu::meta

# ifndef MLN_INCLUDE_ONLY

namespace shape
{

template <typename P>
inline
inertia2d<P>::inertia2d()
{
	init();
}

template <typename P>
inline
void
inertia2d<P>::init(){
	area_ = 0;
	sum_ = literal::zero;
	sum2_ = literal::zero;
}

template <typename P>
inline
void
inertia2d<P>::take(const argument& p_site)
{
	area_++;
	sum_ += p_site.to_vec();
	Ppow  vec = p_site.to_vec();
	for (unsigned i=0; i<Ppow::dim; i++)
		vec[i] *= vec[i];

	sum2_ += vec;
}

template <typename P>
inline
void
inertia2d<P>::take(const inertia2d<P>& other){

	area_ += other.area_;
	sum_ += other.sum_;
	sum2_ += other.sum2_;

}

template <typename P>
inline
float
inertia2d<P>::to_result() const{

	float sum = 0;
	for (unsigned i=0; i<Ppow::dim; i++){
		sum += sum2_[i] - sum_[i]*sum_[i]/area_;
	}

	return (sum/pow(area_,(1+2/Ppow::dim)));
}

template <typename P>
inline
bool
inertia2d<P>::is_valid() const{
	return true;
}

} // end of namespace mln::accu::shape

# endif // ! MLN_INCLUDE_ONLY

} // end of namespace mln::accu

} // end of namespace mln


#endif // ! MLN_ACCU_SHAPE_INERTIA2D_HH

