/*
 * deviation.hh
 *
 *  Created on: May 12, 2010
 *      Author: mauro
 */

#ifndef MLN_MORPHO_ATTRIBUTE_DEVIATION_HH
#define MLN_MORPHO_ATTRIBUTE_DEVIATION_HH

#include <mln/accu/stat/variance.hh>

/// \file
///
/// Define an accumulator that computes the standard deviation, variance, mean and sum.
/// Same as mln::morpho::attribute::variance but returning the value of standard deviation instead of the variance.

namespace mln{

// Forward declaration.
namespace morpho {
namespace attribute {
template <typename T,
typename S = mln_sum(mln_value(T)),
typename R = S>
class deviation;
}
}


// Traits.

namespace trait{

template <typename T, typename S, typename R>
//template <typename T,
//	typename S = mln_sum(mln_value(T)),
//	typename R = S>
struct accumulator_< morpho::attribute::deviation<T,S,R> >
{
	typedef accumulator::has_untake::no    	has_untake;
	typedef accumulator::has_set_value::no 	has_set_value;
	typedef accumulator::has_stop::no      	has_stop;
	typedef accumulator::when_pix::use_v	when_pix;		// Related to the shape of the components
};

} // end of namespace mln::trait

namespace morpho
{

namespace attribute
{

/// deviation accumulator class.
///
/// The parameter \p T is the image type on which the accumulator
/// of values is built.
template <typename T, typename S, typename R>
struct deviation : public mln::accu::internal::base< R , deviation<T,S,R> >
{
	typedef mln_value(T) argument;
	typedef R result;

	deviation();

	/// Manipulators.
	/// \{
	void init();
	void take_as_init_(const argument& t);
	void take(const argument& t);
	void take(const deviation<T,S,R>& other);

	void take(unsigned n_times, const argument& t); // Extra.
	/// \}

	/// Get the accumulator result (the deviation value).
	R to_result() const;

	/// Get the deviation value.
	R var() const;

	/// Get the standard deviation value.
	R standard_deviation() const;

	/// Get the mean value.
	R mean() const;

	/// Get the sum value.
	S sum() const;

	/// Get the number of items.
	unsigned n_items() const;

	/// Check whether this accu is able to return a result.
	/// Always true here.
	bool is_valid() const;

protected:
	typename mln::accu::stat::variance<mln_value(T), S, R> var_;
};

# ifndef MLN_INCLUDE_ONLY

template <typename T, typename S, typename R>
inline
deviation<T,S,R>::deviation()
{
	init();
}

template <typename T, typename S, typename R>
inline
void
deviation<T,S,R>::init()
{
	var_.init();
}

template <typename T, typename S, typename R>
inline
void
deviation<T,S,R>::take(const argument& t)
{
	if (! is_valid())
	{
		take_as_init_(t);
		return;
	}

	var_.take(t);
}

template <typename T, typename S, typename R>
inline
void
deviation<T,S,R>::take(unsigned n_times, const argument& t)
{
	if (! is_valid())
	{
		take_as_init_(t);
		return;
	}

	var_.take(n_times, t);

}

template <typename T, typename S, typename R>
inline
void
deviation<T,S,R>::take_as_init_(const argument& t)
{
	var_.take_as_init_(t);

}


template <typename T, typename S, typename R>
inline
void
deviation<T,S,R>::take(const deviation<T,S,R>& other)
{
	mln_invariant(is_valid());
	var_.take(other.var_);
}

template <typename T, typename S, typename R>
inline
R
deviation<T,S,R>::to_result() const
{
	return (var_.standard_deviation());
}

template <typename T, typename S, typename R>
inline
R
deviation<T,S,R>::mean() const
{
	return (var_.mean());
}

template <typename T, typename S, typename R>
inline
S
deviation<T,S,R>::sum() const
{
	return (var_.sum());
}

template <typename T, typename S, typename R>
inline
unsigned
deviation<T,S,R>::n_items() const
{
	return (var_.n_items());
}

template <typename T, typename S, typename R>
inline
R
deviation<T,S,R>::var() const
{
	return (var_.var());
}

template <typename T, typename S, typename R>
inline
R
deviation<T,S,R>::standard_deviation() const
{
	return (var_.standard_deviation());
}

template <typename T, typename S, typename R>
inline
bool
deviation<T,S,R>::is_valid() const
{
	return var_.is_valid();
}

# endif // ! MLN_INCLUDE_ONLY


} // end of namespace mln::morpho::attribute

} // end of namespace mln::morpho

} // end of namespace mln
#endif // ! MLN_MORPHO_ATTRIBUTE_DEVIATION_HH
