/*
 * sub.hh
 *
 *  Created on: May 6, 2010
 *      Author: mauro
 *      $Id: sub.hh 14 2010-09-27 09:20:36Z mauro $
 */

#ifndef MLN_MORPHO_TREE_FILTER_SUB_HH
#define MLN_MORPHO_TREE_FILTER_SUB_HH

# include <mln/core/concept/function.hh>
# include <mln/fun/ops.hh>

# include <mln/morpho/tree/data.hh>
# include <mln/morpho/tree/propagate_if.hh>

/**
 ** \file   mln/morpho/tree/filter/subtractive.hh
 **
 ** \brief  Filtering with subtractive strategy.
 **
 ** Different from mln::morpho::tree::filter::subtractive.
 ** Implement the subtractive pruning strategy as defined in [1].
 **
 ** [1]	Urbach, E. R.; Roerdink, J. B. T. M. & Wilkinson, M. H. F.
 ** 	"Connected Shape-Size Pattern Spectra for Rotation and Scale-Invariant Classification of Gray-Scale Images"
 ** 	IEEE Transactions on Pattern Analysis and Machine Intelligence, 2007, 29, 272-285.
 **
 **
 */

namespace mln
{

namespace morpho
{

namespace tree
{

namespace filter
{

/**
 ** Subtractive pruning strategy. The node is removed if it
 ** does not verify the predicate. The sub-components values
 ** are lowered by the value of the difference between the value
 ** of the node and the value of its parent.
 **
 ** \param[in] tree Component tree.
 ** \param[out] f_ Image to filter.
 ** \param[in] pred_ Filtering criterion.
 */
template <typename T, typename F, typename P>
inline
void
sub(const T& tree, Image<F>& f_, const Function_v2b<P>& pred_);




# ifndef MLN_INCLUDE_ONLY


template <typename T, typename F>
inline
void
propagate_to_descendants(const T& tree, Image<F>& f_, mln_psite(F) n, mln_value(F) v)
{
	F& f = exact(f_);
	f(n) -= v;
	// propagate the effect to all the children of node n
	p_array<mln_psite(F)> children = tree.children(n);
	mln_piter(p_array<mln_psite(F)>) p(children);

//	 p_array<typename F::psite> children = tree.children(n);
//	 mln_piter(typename F::psite) p(children);
	 for_all(p){
		 propagate_to_descendants(tree, f, p, v);
	 }
}

template <typename T, typename F, typename P>
inline
void
sub(const T& tree, Image<F>& f_, const Function_v2b<P>& pred_)
{
	F& f = exact(f_);
	const P& pred = exact(pred_);

	trace::entering("mln::morpho::tree::filter::subtractive");

//	morpho::tree::propagate_if(tree, f, morpho::tree::desc_propagation (), !pred);

	// Loop on the nodes of the tree
	mln_up_node_piter(T) n(tree);
	for_all(n){
//		std::cout << "- Process node: " << n << " Value: " << (double)f(n) << std::endl;

		if (!pred(n)){

			mln_value(F) v = f(n)-f(tree.parent(n));
			propagate_to_descendants(tree, f, n, v);

//			mln_dn_node_piter(T) descendant(tree);
//
//			bool start = false;
//			// Loop on only the descendants of node n
//			for_all(descendant){
//				if (n == descendant)
//					start = true;
//				if (start && !pred(n)){
//					std::cout << "\tProcess descendant: " << descendant << " ";
//					// if the node is removed, lower the value of all
//					// its descendants by the value of the node
//					std::cout << " Previous value: " << (double)f(descendant) << " ";
//					f(descendant) -= f(n)-f(tree.parent(n));
//					std::cout << " New value: " << (double)f(descendant) << std::endl;
//				}
//			}
//
//			std::cout << "- End Process node: " << n << std::endl;
		}
	}

	trace::exiting("mln::morpho::tree::filter::subtractive");
}

# endif // ! MLN_INCLUDE_ONLY

} // end of namespace mln::morpho::tree::filter
} // end of namespace mln::morpho::tree
} // end of namespace mln::morpho
} // end of namespace mln

#endif // ! MLN_MORPHO_TREE_FILTER_SUB_HH
