// Matlab header
#include "mex.h"
#include "matrix.h"

#include <mln/data/sort_psites.hh>
#include <mln/morpho/tree/data.hh>
#include <mln/morpho/tree/compute_attribute_image.hh>
#include <mln/morpho/attribute/card.hh>
#include <mln/morpho/attribute/volume.hh>

#include <mln/morpho/tree/filter/all.hh>
#include <mln/morpho/tree/propagate_representative.hh>

#include <mln/morpho/closing/area.hh>
#include <mln/morpho/opening/area.hh>

#include <iostream>
#include <vector>

#include "mexmln_glue_import.hpp"
#include "mex-mln_glue_attribute_profile.hpp"

// da shell: 	g++-4.1 -c maxtree.cpp data.cpp -fPIC -I/home/mauro/install/lnk_C++Lib/boost/boost_1_39_0
// in matlab:	mex main.cpp data.o maxtree.o -I/home/mauro/install/C++Lib/boost/boost_1_42_0 -I/home/mauro/install/C++Lib/olena-1.0/include
//				mex main.cpp ../Debug/src/filter.o -I/home/mauro/install/C++Lib/boost/boost_1_42_0 -I/home/mauro/install/C++Lib/olena-1.0/include


/* TODO
 *  - sistemare check all'inizio
 *  - generalizzare funzione di mex2mln
 *  - eventualmente estrarre calcolo tree da algoritmo attribute_filter
 *  ----------------
 */


/*
 * Gateway Function
 * Interface between Matlab and the C++ routine.
 *
 * Input:
 * 	Immagine			prhs[0]
 * 	Tipo Attributo		prhs[1]
 * 	Lambdas				prhs[2]
 * 	Connectivity		prhs[3]
 *  Filtering rule      prhs[4]
 *
 * Output:
 * 	Array di immagini	plhs[0]
 */
void mexFunction(int nlhs, 			// Number of outputs
mxArray *plhs[],			// Array containing the outputs
int nrhs, 					// Number of inputs
const mxArray *prhs[]){		// Array containing the inputs
    
     /************************************************
      *              Input checks
      ************************************************/
    
    if ((nrhs > 5) || (nrhs == 1) || (nrhs == 2)){
        mexErrMsgTxt("Wrong number of inputs.\n");
    }
    if (nrhs == 0){
        mexPrintf(" Attribute Profile - Sinopsys\n");
        mexPrintf("\n");
        mexPrintf("   AP = ATTRIBUTE_PROFILE(I, A, L)\n");
        mexPrintf("   AP = ATTRIBUTE_PROFILE(I, A, L, C)\n");
        mexPrintf("   AP = ATTRIBUTE_PROFILE(I, A, L, C, F)\n");
        mexPrintf("\n");
        mexPrintf(" INPUT\n");
        mexPrintf("     I:  2D input image. Data type supported: uint8, int8, uint16, int16, uint32, int32, uint64, int64.\n");
        mexPrintf("     A:  Attribute type. String.\n");
        mexPrintf("         Attribute supported:\n");
        mexPrintf("             Area: 'a', 'area';\n");
        mexPrintf("             Length of the diagonal of the bounding box: 'd', 'diagonal';\n");
        mexPrintf("             Moment of inertia: 'i', 'inertia';\n");
        mexPrintf("             Standard deviation: 's', 'std'.\n");
        mexPrintf("     L:  Values of lambda. Array of double.\n");
        mexPrintf("     C:  Connectivity. Scalar. Either 4 (default) or 8.\n");
        mexPrintf("     F:  Filtering rule. String.\n");
        mexPrintf("         Rules supported:\n");
        mexPrintf("             Min: 'min' (default for attributes 'a', 'd');\n");
        mexPrintf("             Max: 'max';\n");
        mexPrintf("             Direct: 'direct';\n");
        mexPrintf("             Sub: 'sub' (default for attributes 'i', 's').\n");
        mexPrintf("\n");
        mexPrintf(" OUTPUT\n");
        mexPrintf("     AP: Attribute Profile. Array of 2D images of the same data type of the input image.\n");
        mexPrintf("\n");
        mexPrintf("--\n");
        mexPrintf(" Version 1.0\n");
        mexPrintf(" 14 May 2010\n");
        mexPrintf(" Mauro Dalla Mura\n");
        mexPrintf(" dallamura@disi.unitn.it\n");
return;
    }
    
    // Set default values
    std::string filter_rule;    // In the internal routine is set the real default value (according to the attribute type).
    int conn = 0;   // In the internal routine is set the real default value (according to the attribute type).
    
    // Check the inputs.
    if (nrhs > 4){
        // Filtering rule
        if (!mxIsEmpty(prhs[4])){
            if (!mxIsChar(prhs[4]))
                mexErrMsgTxt("Input 5 has to be either a string or empty.");
            std::string tmp(mxArrayToString(prhs[4]));
            filter_rule = tmp;
            mexPrintf("Filtering Rule: %s\n", filter_rule.c_str());
        }
    }
    if (nrhs > 3){
        // Connectivity
        if (!mxIsEmpty(prhs[3])){
            
            if (!mxIsDouble(prhs[3]) || mxGetNumberOfElements(prhs[3]) != 1)
                mexErrMsgTxt("Input 4 has to be either a scalar double or empty.");
            double* in_ptr = (double*)mxGetPr(prhs[3]);
            conn = static_cast<int>(in_ptr[0]);
        }
    }
    
    //	prhs[0] - Input image
    if (mxIsEmpty(prhs[0]) ||
    (!mxIsNumeric(prhs[0]) &&
    !mxIsLogical(prhs[0])))
        mexErrMsgTxt("Input 1 has to be a numeric 1D or 2D array.");
    if (mxGetNumberOfDimensions(prhs[0]) > 2)
        mexErrMsgTxt("Input 1. Data with more than two dimensions are not supported.");
    
    //	prhs[0] - Attribute type
    if (mxIsEmpty(prhs[1]) || !mxIsChar(prhs[1]))
        mexErrMsgTxt("Input 2 has to be a string.");
    
    std::string attr_type(mxArrayToString(prhs[1]));
    
    //	Input 2
    if (mxIsEmpty(prhs[2]))
        mexErrMsgTxt("Input 3 has not to be empty.");   
    
    /************************************************
     *          End of Input checks
     ************************************************/
    
    //  Call the routine
    mxClassID   category;
    category = mxGetClassID(prhs[0]);
    switch (category)  {
        
        // FIXME: There an issue with the variance attribute when it is instantiated with booleans.
//                 case mxLOGICAL_CLASS: compute_attribute_profile<bool>(prhs[0], plhs[0], attr_type, conn, prhs[2], filter_rule);    break;      // MATLAB logical
        case mxINT8_CLASS:  compute_attribute_profile<signed char>(prhs[0], plhs[0], attr_type, conn, prhs[2], filter_rule);    break;      // MATLAB int8
        case mxUINT8_CLASS:  compute_attribute_profile<unsigned char>(prhs[0], plhs[0], attr_type, conn, prhs[2], filter_rule); break;  // MATLAB uint8
        case mxINT16_CLASS:  compute_attribute_profile<signed short>(prhs[0], plhs[0], attr_type, conn, prhs[2], filter_rule);  break;      // MATLAB int16
        case mxUINT16_CLASS: compute_attribute_profile<unsigned short>(prhs[0], plhs[0], attr_type, conn, prhs[2], filter_rule);  break;      // MATLAB uint16
        case mxINT32_CLASS:  compute_attribute_profile<signed int>(prhs[0], plhs[0], attr_type, conn, prhs[2], filter_rule);   break;      // MATLAB int32
        case mxUINT32_CLASS: compute_attribute_profile<unsigned int>(prhs[0], plhs[0], attr_type, conn, prhs[2], filter_rule);   break;      // MATLAB uint32
        case mxINT64_CLASS:  compute_attribute_profile<signed long>(prhs[0], plhs[0], attr_type, conn, prhs[2], filter_rule);   break;      // MATLAB int64
        case mxUINT64_CLASS: compute_attribute_profile<unsigned long>(prhs[0], plhs[0], attr_type, conn, prhs[2], filter_rule);   break;      // MATLAB uint64
        //         case mxSINGLE_CLASS:
        //             mexErrMsgTxt("Data type not supported.\n Please convert the input data to either one of the following types: uint8, int8, uint16, int16, uint32, int32..."); break;
        //         case mxDOUBLE_CLASS:
        //                         mexErrMsgTxt("Data type not supported.\n Please convert the input data to either one of the following types: uint8, int8, uint16, int16, uint32, int32..."); break;
        default: mexErrMsgTxt("Data type not supported.\n Please convert the input data to either one of the following types: uint8, int8, uint16, int16, uint32, int32, uint64, int64."); break;
    }
}
