// Matlab header
#include "mex.h"
#include "matrix.h"

#include <mln/core/image/image2d.hh>
#include <mln/core/alias/neighb2d.hh>
#include <mln/core/var.hh>
#include <mln/pw/all.hh>

#include <mln/data/sort_psites.hh>
#include <mln/morpho/tree/data.hh>
#include <mln/morpho/tree/compute_attribute_image.hh>
#include <mln/morpho/attribute/card.hh>
#include <mln/morpho/attribute/volume.hh>

#include <mln/morpho/tree/filter/all.hh>
#include <mln/morpho/tree/propagate_representative.hh>

#include <mln/morpho/closing/area.hh>
#include <mln/morpho/opening/area.hh>

#include "mln/morpho/attribute/diagonal.hh"
#include "mln/morpho/attribute/inertia.hh"
#include "mln/morpho/attribute/deviation.hh"
#include "mln/morpho/tree/filter/sub.hh"


#include <iostream>
#include <vector>
#include <map>
#include <string>

#define VERBOSE false


// Resolve type of image (e.g., 2d, 3d, etc.) and connectivity
template <typename Tmex>        // Tmex has to be a built in type.
void compute_attribute_profile(const mxArray* image,
mxArray*& attr_profile,
const std::string attr_type,
int conn,
const mxArray* lambda,
std::string filter_rule){
    unsigned int ndims = mxGetNumberOfDimensions(image);
    switch(ndims){
        case 2:{
            typedef mln::image2d<Tmex> I;
            
            // Read the image
            I image_resolved = mxArray2mln_image<Tmex, Tmex>(image);
            
            // Create the output array MOVE IT TO THE TEMPLATE BELOW???
            unsigned long m = image_resolved.nrows();
            unsigned long n = image_resolved.ncols();
            unsigned int nlevels = mxGetN(lambda);  // ATH! we are accessing a variable which is not completely checked here
            
            const mwSize dims[]={m, n, 2*nlevels+1};
            attr_profile = mxCreateNumericArray(
            3, 					// Number of dimension
            dims, 				// Dimensions size
            define_image_type(Tmex(0)),// Use the same data type of the input
            mxREAL);			// Real data (non complex)
            
            Tmex* pout= (Tmex*)mxGetPr(attr_profile);		// cast the type of the ptr according to the input data type

            // Resolve connectivity for 2d images
            switch(conn){
                case 0: conn = 4; // Default value 4-conn
                case 4:
                    compute_attribute_profile(image_resolved, pout, attr_type, mln::c4(), lambda, filter_rule);
                    if (VERBOSE)
                        mexPrintf("Connectivity: %d\n", conn);
                    break;
                case 8: 
                    compute_attribute_profile(image_resolved, pout, attr_type, mln::c8(), lambda, filter_rule); 
                     if (VERBOSE)
                        mexPrintf("Connectivity: %d\n", conn);
                    break;
                
                //                 default:  mexErrMsgTxt("The value of connectivity, %f, does not match any connectivity rule for 2d images.", conn);
                default:  mexErrMsgTxt("This value of connectivity does not match any connectivity rule for 2d images.");
                
            }
            break;
        }
        
        //         default:  mexErrMsgTxt("Images of %d dimensions are not supported.", ndims);
        default:  mexErrMsgTxt("Images of >2 dimensions are not supported.");
        
    }
}

// Resolve attribute type and values of lambda
template <typename I, typename N, typename Tmex>
void compute_attribute_profile(const mln::Image<I>& image,
Tmex*& attr_profile,
// mln::Image<I>::value*& attr_profile,
std::string attr_type,
const mln::Neighborhood<N>& nbh,
const mxArray* lambda,
std::string filter_rule){
    //     switch (attr_type){
    //         case 'a':template <typename T>
    
    
    if (attr_type == "a" ||
    attr_type == "area"){
        
        typedef mln::morpho::attribute::card<I> attr;
        typedef __typeof__(mln_result(attr)) Tattr;
        
        // Check lambdas
        mxClassID   category;
        category = mxGetClassID(lambda);
        switch (category){
            case mxDOUBLE_CLASS:break;
            default: mexErrMsgTxt("Input 3 has to be a 1xN array of double."); break;
        }
        if (mxGetM(lambda) != 1)
            mexErrMsgTxt("Input 3 has to be a 1xN array of double.");
        // End of checks
        
        std::vector<Tattr> lambda_resolved; // Resolve the values of lambda
        
        double* lambda_ptr = (double*)mxGetPr(lambda);
        
        unsigned nlevels = mxGetN(lambda);			// Cols!
        
        for (unsigned i=0; i<nlevels; i++){
            lambda_resolved.push_back( static_cast<Tattr>(lambda_ptr[i]));       // here, perform the cast to the type expected by the attribute.
        }
        if (filter_rule.empty())
            filter_rule = "min";        // Set default filtering rule
        if (VERBOSE){
            mexPrintf("Attribute: area\n");
            mexPrintf("Lambdas: ");
            for (unsigned i=0; i<nlevels; i++){
                mexPrintf("%f ", lambda_resolved.at(i));
            }
            mexPrintf("\n");
            mexPrintf("Filtering Rule: %s\n", filter_rule.c_str());
        }
        _compute_attribute_profile(mln::exact(image), attr_profile, attr(), mln::exact(nbh), lambda_resolved, filter_rule);
    }
    else if(attr_type == "d" ||
    attr_type == "diagonal"){
        
        typedef mln::morpho::attribute::diagonal<I> attr;
        
        // Check lambdas
        mxClassID   category;
        category = mxGetClassID(lambda);
        switch (category){
            case mxDOUBLE_CLASS:break;
            default: mexErrMsgTxt("Input 3 has to be a 1xN array of double."); break;
        }
        if (mxGetM(lambda) != 1)
            mexErrMsgTxt("Input 3 has to be a 1xN array of double.");
        // End of checks
        
        std::vector<__typeof__(mln_result(attr))> lambda_resolved; // Resolve the values of lambda
        
        double* lambda_ptr = (double*)mxGetPr(lambda);
        
        unsigned nlevels = mxGetN(lambda);			// Cols!
        
        for (unsigned i=0; i<nlevels; i++){
            lambda_resolved.push_back( static_cast<__typeof__(mln_result(attr))>(lambda_ptr[i]));       // here, perform the cast to the type expected by the attribute.
        }
        if (filter_rule.empty())
            filter_rule = "min";        // Set default filtering rule
        if (VERBOSE){
            mexPrintf("Attribute: diagonal\n");
            mexPrintf("Lambdas: ");
            for (unsigned i=0; i<nlevels; i++){
                mexPrintf("%f ", lambda_resolved.at(i));
            }
            mexPrintf("\n");
            mexPrintf("Filtering Rule: %s\n", filter_rule.c_str());
        }
        _compute_attribute_profile(mln::exact(image), attr_profile, attr(), mln::exact(nbh), lambda_resolved, filter_rule);
    }
    else if(attr_type == "i" ||
    attr_type == "inertia"){
        
        typedef mln::morpho::attribute::inertia<I> attr;
        
        // Check lambdas
        mxClassID   category;
        category = mxGetClassID(lambda);
        switch (category){
            case mxDOUBLE_CLASS:break;
            default: mexErrMsgTxt("Input 3 has to be a 1xN array of double."); break;
        }
        if (mxGetM(lambda) != 1)
            mexErrMsgTxt("Input 3 has to be a 1xN array of double.");
        // End of checks
        
        std::vector<__typeof__(mln_result(attr))> lambda_resolved; // Resolve the values of lambda
        
        double* lambda_ptr = (double*)mxGetPr(lambda);
        
        unsigned nlevels = mxGetN(lambda);			// Cols!
        
        for (unsigned i=0; i<nlevels; i++){
            lambda_resolved.push_back( static_cast<__typeof__(mln_result(attr))>(lambda_ptr[i]));       // here, perform the cast to the type expected by the attribute.
        }
        if (filter_rule.empty())
            filter_rule = "sub";        // Set default filtering rule
        if (VERBOSE){
            mexPrintf("Attribute: inertia\n");
            mexPrintf("Lambdas: ");
            for (unsigned i=0; i<nlevels; i++){
                mexPrintf("%f ", lambda_resolved.at(i));
            }
            mexPrintf("\n");
            mexPrintf("Filtering Rule: %s\n", filter_rule.c_str());
        }
        _compute_attribute_profile(mln::exact(image), attr_profile, attr(), mln::exact(nbh), lambda_resolved, filter_rule);
    }
    else if(attr_type == "s" ||
    attr_type == "std"){
        
        typedef mln::morpho::attribute::deviation<I> attr;
        
        // Check lambdas
        mxClassID   category;
        category = mxGetClassID(lambda);
        switch (category){
            case mxDOUBLE_CLASS:break;
            default: mexErrMsgTxt("Input 3 has to be a 1xN array of double."); break;
        }
        if (mxGetM(lambda) != 1)
            mexErrMsgTxt("Input 3 has to be a 1xN array of double.");
        // End of checks
        
        std::vector<__typeof__(mln_result(attr))> lambda_resolved; // Resolve the values of lambda
        
        double* lambda_ptr = (double*)mxGetPr(lambda);
        
        unsigned nlevels = mxGetN(lambda);			// Cols!
        
        for (unsigned i=0; i<nlevels; i++){
            lambda_resolved.push_back( static_cast<__typeof__(mln_result(attr))>(lambda_ptr[i]));       // here, perform the cast to the type expected by the attribute.
        }
        if (filter_rule.empty())
            filter_rule = "sub";        // Set default filtering rule
        if (VERBOSE){
            mexPrintf("Attribute: std\n");
            mexPrintf("Lambdas: ");
            for (unsigned i=0; i<nlevels; i++){
                mexPrintf("%f ", lambda_resolved.at(i));
            }
            mexPrintf("\n");
            mexPrintf("Filtering Rule: %s\n", filter_rule.c_str());
        }
        _compute_attribute_profile(mln::exact(image), attr_profile, attr(), mln::exact(nbh), lambda_resolved, filter_rule);
    }
    else
        mexErrMsgTxt("Attribute not supported.");
}


// Core of the algorithm.
template <typename I, typename Tmex, typename A, typename N, typename V>
void _compute_attribute_profile(const I& image,
Tmex*& attr_profile,
const mln::Accumulator<A>& attr_type,
const N& nbh,
std::vector<V> lambda,
std::string filter_rule){
    
    using namespace mln;
    
    // Compute the tree
    typedef p_array< mln_site(I) > S;
    typedef morpho::tree::data<I,S> tree_t;
    
    //  Start with min-tree
    S sorted_sites = data::sort_psites_decreasing(image);	// Decreasing => min-tree
    unsigned long nelem = sorted_sites.nsites();            // Get the number of elements from the site set rather than from the image
    //(because the abstract type of image does not support a mehod returning the number of sites- it is just a funtion)
    //     std::cout << "check the number of elements returned by 'sorted_sites.nsites()': " << nelem << std::endl;
    tree_t tree(image, sorted_sites, nbh);
    
    // Attribute
    
    // typedef the output image with type correspondent to the return type of A
    typedef mln_ch_value(I, __typeof__(mln_result(A))) attr_img;  // = typedef image2d<__typeof__(mln_result(A))> attr_img;
    
    // Compute the attributes on the tree
    attr_img a = morpho::tree::compute_attribute_image(attr_type, tree);
    
//     // Write the image with attributes if it is defined.
//     if (attr_img_out != 0){
//         mln_image2mxPtr<Tattr, Tattr>(a, attr_img_out);
//             attr_img_out += nelem;    // Update the position of pout
//     }
        
    
    // Filter the input image.
    // TODO: Use ITERATORS!
    for (unsigned int i=lambda.size()-1; (i>=0 && i<lambda.size()); i--){	// Inverse Loop on the levels of the profile
        // (i>=0 && i<lambdas.size()) is due to avoid overflow
        
        __typeof__(((pw::value(a)) >= pw::cst(lambda.at(i)))) predicate = ((pw::value(a)) >= pw::cst(lambda.at(i)));
        typedef __typeof__(((pw::value(a)) >= pw::cst(lambda.at(i)))) predicate_t;
        
        I out = duplicate(image);
        if (filter_rule == "min")
            morpho::tree::filter::min(tree, out, predicate);
        else if (filter_rule == "max")
            morpho::tree::filter::max(tree, out, predicate);
        else if (filter_rule == "direct")
            morpho::tree::filter::direct(tree, out, predicate);
        else if (filter_rule == "sub")
            morpho::tree::filter::sub(tree, out, predicate);
        else
            mexErrMsgTxt("The filtering rule must be one of the following strings: 'min', 'max', 'direct', 'sub'.\n");
            
            morpho::tree::propagate_representative(tree, out);
            
            // Write the data to the memory pointed by ptr.
            mln_image2mxPtr<Tmex, Tmex>(out, attr_profile);
            attr_profile += nelem;    // Update the position of pout
    }
    
    //	Add original image
    I out = duplicate(image);
    mln_image2mxPtr<Tmex, Tmex>(out, attr_profile);
    attr_profile += nelem;    // Update the position of pout
    
    //  Max-tree
    sorted_sites = data::sort_psites_increasing(image);	// Increasing => max-tree
    tree_t max_tree(image, sorted_sites, nbh);
    
    a = morpho::tree::compute_attribute_image(attr_type, max_tree);
    
    for (unsigned int i=0; i<lambda.size(); i++){	// Inverse Loop on the levels of the profile
        __typeof__(((pw::value(a)) >= pw::cst(lambda.at(i)))) predicate = ((pw::value(a)) >= pw::cst(lambda.at(i)));
        
        I out = duplicate(image);
        if (filter_rule == "min")
            morpho::tree::filter::min(max_tree, out, predicate);
        else if (filter_rule == "max")
            morpho::tree::filter::max(max_tree, out, predicate);
        else if (filter_rule == "direct")
            morpho::tree::filter::direct(max_tree, out, predicate);
        else if (filter_rule == "sub")
            morpho::tree::filter::sub(max_tree, out, predicate);
        else
            mexErrMsgTxt("The filtering rule must be one of the following strings: 'min', 'max', 'direct', 'sub'.\n");
            morpho::tree::propagate_representative(max_tree, out);
            
            mln_image2mxPtr<Tmex, Tmex>(out, attr_profile);
            attr_profile += nelem;    // Update the position of pout
    }
    
}
