// Matlab header
#include "mex.h"
#include "matrix.h"

#include <mln/core/image/image2d.hh>
#include <mln/core/alias/neighb2d.hh>
#include <mln/core/var.hh>
#include <mln/pw/all.hh>

// Data values
#include <mln/value/int_u8.hh>
#include <mln/value/int_s8.hh>
#include <mln/value/int_u16.hh>
#include <mln/value/int_s16.hh>
#include <mln/value/int_u32.hh>
#include <mln/value/int_s32.hh>

/*
 *  Utility. Template functions that returns the data type as defined in Matlab mex from the C++ type of the data.
 *  Template specialization is employed for getting the correspondent value.
 */
template<typename T> mxClassID define_image_type(T){}
template<> mxClassID define_image_type(bool){return mxLOGICAL_CLASS;}
template<> mxClassID define_image_type(signed char){return mxINT8_CLASS;}
template<> mxClassID define_image_type(unsigned char){return mxUINT8_CLASS;}
template<> mxClassID define_image_type(signed short){return mxINT16_CLASS;}
template<> mxClassID define_image_type(unsigned short){return mxUINT16_CLASS;}
template<> mxClassID define_image_type(signed int){return mxINT32_CLASS;}
template<> mxClassID define_image_type(unsigned int){return mxUINT32_CLASS;}
template<> mxClassID define_image_type(signed long){return mxINT64_CLASS;}
template<> mxClassID define_image_type(unsigned long){return mxUINT64_CLASS;}
template<> mxClassID define_image_type(float){return mxSINGLE_CLASS;}
template<> mxClassID define_image_type(double){return mxDOUBLE_CLASS;}

/*
 *  Utility. Given the type of Site values in mln and the type of data pointed by ptr, 
 *  write in the data from the mln image.
 *  ATH! the data are written according to the Matlab indexing convention (columns, rows).
 */
template <typename Tmln, typename Tmex>
void mln_image2mxPtr(const mln::image2d<Tmln>& input, Tmex*& ptr){

    //	Put data in the Matlab array.
    for (unsigned int j=0; j<input.ncols(); j++)
        for (unsigned int i=0; i<input.nrows(); i++){
            ptr[j*input.nrows()+i] = mln::opt::at(input, i, j);
        }    
}

/*
 *  Utility. Given the type of Site values in mln and the type of data in the mex array, 
 *  allocate the memory in the mex array and copy in the data from the mln image.
 */
template <typename Tmln, typename Tmex>
void mln_image2mxArray(const mln::image2d<Tmln>& input, mxArray*& plhs){
    
    unsigned long m = input.nrows();
    unsigned long n = input.ncols();
    // Create the output data
    const mwSize dims[]={m, n};
    plhs = mxCreateNumericArray(
    2, 					// Number of dimension
    dims, 				// Dimensions size
    define_image_type(Tmex(0)),// Use the same data type of the input
    mxREAL);			// Real data (non complex)
    
    Tmex* pind= (Tmex*)mxGetPr(plhs);		// cast the type of the ptr according to the input data type
    
    // Write the data to the memory pointed by ptr.
     mln_image2mxPtr<Tmln, Tmex>(input, pind);  
}

/*
 *  Utility. Given the type of Site values in mln and the type of data in the mex array, 
 *  return a mln image initialized with the data of the mex array.
 */
template <typename Tmln, typename Tmex>
mln::image2d<Tmln> mxArray2mln_image(const mxArray* prhs){
    
        // Read number of rows and cols.
    unsigned long m = mxGetM(prhs);					// Rows
    unsigned long n = mxGetN(prhs);					// Cols
    
    Tmex* in_image = (Tmex*)mxGetPr(prhs);	// Ptr to the input image
    
    // Write the data to the mln image
    typedef mln::image2d<Tmln> I;
    I input(mln::make::box2d(m, n));		// 	create the image
    
    // Copy the data from the MATLAB input to the mln image.
    for (unsigned long i=0; i<m; i++){
        for (unsigned long j=0; j<n; j++){
            mln::opt::at(input, i, j) = in_image[j*m+i];
        }
    }
    
    return input;
}


/*
 *  Trial. Routine that reads the elements of an image stored in a mxArray and stores them in a mln image and put them back to another mxArray.
 *  The data types are defined by the template call.
 */
template <typename Tmln, typename Tmex>
void import(const mxArray* prhs, mxArray*& plhs){
    
    // Read the image
    mln::image2d<Tmln> input = mxArray2mln_image<Tmln, Tmex>(prhs);
    
    // Return the image
    mln_image2mxArray<Tmln, Tmex>(input, plhs);
    
}



